<?php
/**
 * @author Michele Andreoli <michi.andreoli@gmail.com>
 * @name CSSUnifier.class.php
 * @version 0.1 updated 10-08-2011
 * @license http://opensource.org/licenses/gpl-license-php GNU Public License
 * @package CSSUnifier
 */

/**
 * CSS compressor and unifier class
 */
class Unifier {
    private $css = "aggregate_min";
    private $cssDir = null;
    private $cssList = array();
    private $ext = ".css";
    private $sub = "_min";
    private $props = array();
    private $log = null;   

    /**
     * Constructor for CUnifier
     * @param $xml string XML file path
     * @param [$path] string Path of the logger file
     */
    public function __construct ($xml, $log=null) {
        if ($log) {
            $this->log = $log;
            $this->deleteLog();
        }
        
        try {
            $this->cssList = $this->readXML($xml);
            $this->copyCSS();
        }
        catch (UnifierException $e) {
            echo $e;
            $this->logging("FAILURE! {$e}");
            exit(0);
        }        
    }

    /**
     * Get the string for CSS request of your page:
     * <link rel="stylesheet" type="text/css" href="<path>"/>
     * @return string
     */
    public function getCSS() {
        $this->logging("=> GENERATED OUTPUT");
        if ($this->props['unifier'] == "true") {
            $this->logging("<link rel=\"stylesheet\" type=\"text/css\" href=\"{$this->css}\"/>");
            echo "<link rel=\"stylesheet\" type=\"text/css\" href=\"{$this->css}\"/>";
        }
        else {
            foreach($this->cssList as $file) {
                 $this->logging("<link rel=\"stylesheet\" type=\"text/css\" href=\"{$file['tmpPath']}\"/>");
                 echo "<link rel=\"stylesheet\" type=\"text/css\" href=\"{$file['tmpPath']}\"/>";
            }
        }
    }

    /**
     * Read XML document
     * @param $xml string XML file path
     * @return array[]
     */
    private function readXML($xml) {
        $this->logging("=> STARTING TO PROCESS XML FILE '{$xml}'");
        if (!file_exists($xml)) {
            throw new UnifierException("the XML file '{$xml}' does not exist! Check the path.");
        }

        if (!($tmp = simplexml_load_file($xml))) {
            throw new UnifierException("impossible to read XML '{$xml}'. Check the permissions of this file or his syntax.");
        }
        $biTmp = array();

        //Get properties
        $this->logging("Reading mandatories properties...");
        foreach ($tmp->properties->property as $elem) {
            foreach ($elem->attributes() as $key => $value) {
                $this->props[$key] = $value;
            }
        }

        //Check for properties
        if (count($this->props) != 4) {
            throw new UnifierException("missing properties in XML file '{$xml}'.");
        }
        $this->logging("OK!");
        $this->css = $this->props['path'].$this->sub."/".$this->props['outname'].$this->ext;       

        //Get css
        foreach ($tmp->csslist->css as $elem) {
            $properties = array(
                 'path'      => $elem['path'].$elem['name'].$this->ext
                ,'tmpPath'   => $this->props['path'].$this->sub."/".$elem['name'].$this->sub.$this->ext
                ,'name'      => $elem['name']
                ,'tmpName'   => $elem['name'].$this->sub
                ,'compress'  => $elem['compress']
                ,'outOfDate' => "false"
            );            

            //If the original css missing or his length is 0 don't push into array
            if (file_exists($properties['path']) && filesize($properties['path']) > 0) {
                array_push($biTmp, $properties);
                //Make the path of the output directory for copied css
                $this->cssDir = $this->props['path'].$this->sub."/";
            }
            else {
                throw new UnifierException("this css file '{$properties['path']}' specified in the XML '{$xml}' does not exist or is not readable! Remove this path from XML or check the permissions of this file and his parents directories.");
            }

            $this->logging("Reading attributes for '{$properties['path']}'...");
            //Check for attributes
            if ($elem['path'] == "" || $elem['name'] == "" || $elem['compress'] == "") {
                throw new UnifierException("missing css attributes in XML file '{$xml}'.");
            }
            $this->logging("OK!");
        }        

        //Check for a path
        if ($this->cssDir == "") {
            throw new UnifierException("There aren't css files specified into XML file '{$xml}'.");
        }

        if ($this->props['forceRefresh'] == "true") {
            try {
                $this->deleteDirectory($this->cssDir);
            }
            catch (Exception $e){
                throw new UnifierException($e->getMessage());
            }
        }
        $this->logging("=> END TO PROCESS XML FILE '{$xml}'\n");

        return $biTmp;
    }

    /**
     * Create a copy of css files if not exists and sync the copies
     * @return bool
     */
    private function copyCSS() {
        $this->logging("=> STARTING TO PROCESS INPUT CSS FILES");
        //If there isn't copy directory
        if (!file_exists($this->cssDir)) {
            //Create copy directory
            $this->logging("Creating temporary directory '{$this->cssDir}'...");
            if (!mkdir($this->cssDir, 0755)) {
                throw new UnifierException("impossible to create directory '{$this->cssDir}'. Check the permissions of the parents directories.");
            }
            $this->logging("OK!");
        }        

        foreach($this->cssList as &$file) {
            $this->logging("==> STARTING TO PROCESS CSS FILE '{$file['tmpPath']}'");
            //If there isn't file's copy
            if (!file_exists($file['tmpPath']) || $this->isOutOfDate($file['tmpPath'], $file['path'])) {
                //Copy css files
                $this->logging("Copying '{$file['path']}' to '{$file['tmpPath']}'...");
                if (!copy($file['path'], $file['tmpPath'])) {
                    throw new UnifierException("impossible to copy file '{$file['path']}' in file '{$file['tmpPath']}'. Check the permissions of directory {$this->cssDir}.");
                }                
                chmod($file['tmpPath'], 0755);
                $this->logging("OK!");

                //Sync the copies
                $file['outOfDate'] = "true";                
                try {
                    $this->setUpToDate($file['path'], $file['tmpPath']);
                }
                catch (Exception $e) {
                    throw new UnifierException($e->getMessage());
                }
            }
            $this->logging("==> END TO PROCESS CSS FILE\n");
        }
        $this->logging("=> END TO PROCESS INPUT CSS FILES\n");
        
        $this->logging("=>STARTING TO CREATE OUTPUT CSS FILES");
        //Create ouput css if not exist or if isn't up to date
        if (!file_exists($this->css) || $this->isCSSOutOfDate($this->css)) {
            if ($this->props['unifier'] == "true") {
                $this->logging("Creating aggregate css '{$this->css}'...");
                if (!$fp = fopen($this->css, 'w')) {
                    throw new UnifierException("impossible to create file '{$this->css}'. Check the permissions of this file or his parents directories.");
                }                
                chmod($this->css, 0755);
                $this->logging("OK!");
                fclose($fp);
            }

            foreach($this->cssList as &$file) {
                $this->logging("==> STARTING TO PROCESS CSS FILE '{$file['tmpPath']}'");
                $context = null;

                //Read content of the current css
                $this->logging("Reading css '{$file['tmpPath']}'...");
                if (!$fp = fopen($file['tmpPath'], 'r')) {
                    throw new UnifierException("impossible to read file '{$file['tmpPath']}'. Check the permissions of this file or his parents directories.");
                }               
                $context = fread($fp, filesize($file['tmpPath']));
                $this->logging("OK!");
                fclose($fp);

                //If file isn't up to date, compress it before append
                if ($file['compress'] == "true" && $file['outOfDate'] == "true") {
                    $this->logging("Minimizing '{$file['tmpPath']}'...");
                    $context = $this->minify($context);                    

                    //Replace the content with the minify version
                    $this->logging("Replacing content of '{$file['tmpPath']}' with minify version...");
                    if (!$fp = fopen($file['tmpPath'], 'w')) {
                        throw new UnifierException("impossible to write file '{$file['tmpPath']}'. Check the permissions of this file or his parents directories.");
                    }
                    fwrite($fp, $context);
                    $this->logging("OK!");
                    fclose($fp);

                    //Re-sync the modified file
                    $file['outOfDate'] = "false";
                    try {
                        $this->logging("Resynchronizing...");
                        $this->setUpToDate($file['path'], $file['tmpPath']);
                    }
                    catch (Exception $e) {
                        throw new UnifierException($e->getMessage());
                    }
                }                
                
                //Append content of the current css in the output css
                if ($this->props['unifier'] == "true") {
                    $this->logging("Appending to aggregate css '{$this->css}'...");
                    if (!$fp = fopen($this->css, 'a')) {
                        throw new UnifierException("impossible to append into file '{$this->css}'. Check the permissions of this file or his parents directories.");
                    }
                    fwrite($fp, $context);
                    $this->logging("OK!");
                    fclose($fp);
                }
                $this->logging("==> END TO PROCESS CSS FILE\n");
            }
            
            //Sync the output css
            if ($this->props['unifier'] == "true") {
                try {
                    $this->logging("Resynchronizing...");
                    $this->setCSSUpToDate($this->css);
                }
                catch (Exception $e) {
                    throw new UnifierException($e->getMessage());
                }
            }
        }
        $this->logging("=>END TO CREATE OUTPUT CSS FILES\n");

        return true;
    }

    /**
     * Check if the file is out of date
     * @param $file1 string Filename to check
     * @param $file2 string Filename to check with the first one
     * @return bool
     */
    private function isOutOfDate($file1, $file2) {
        $last = filemtime($file1);
        $current = filemtime($file2);

        //If timestamps aren't equal, the copy is out of date
        if ($last != $current) {
            $this->logging("'{$file1}' is out of date!");
            return true;
        }
        $this->logging("'{$file1}' is up to date!");

        return false;
    }

    /**
     * Set all files up to date
     * @param $file1 string Filename
     * @param $fileToUpdate string Filename to update
     * @return bool
     */
    private function setUpToDate($file1, $file2) {
        $file = filemtime($file1);
        $update = $file2;


        $this->logging("Updating '{$file2}' with timestamp {$file}...");
        if (!touch($update, $file)) {
            throw new UnifierException("impossible to update file '{$update}'. Check the permissions of this file.");
        }
        $this->logging("OK!");

        return true;
    }

    /**
     * Check if the output file is out of date
     * @param $file1 string Filename to check
     * @return bool
     */
    private function isCSSOutOfDate($css) {
        $output = filemtime($css);

        foreach ($this->cssList as $file) {
            if (filemtime($file['tmpPath']) > $output) {
                $this->logging("'{$css}' is out of date!");
                return true;
            }
        }
        $this->logging("'{$css}' is up to date!");

        return false;
    }

    /**
     * Set output css up to date
     * @param $file1 string Filename to update
     * @return bool
     */
    private function setCSSUpToDate($css) {
        $max = 0;
        $name = null;

        foreach ($this->cssList as $file) {
            $current = filemtime($file['tmpPath']);
            if ($current > $max) {
                $max = $current;
                $name = $file['tmpPath'];
            }
        }

        $this->logging("Updating '{$css}' with timestamp {$max} of the last modified file '{$name}'...");
        if (!touch($css, $max)) {
            throw new UnifierException("impossible to update file '{$css}'. Check the permissions of this file.");
        }
        $this->logging("OK!");
        
        return true;
    }

    /**
     * Minify the content of css
     * @param $context string The content of css to compress
     * @return string
     */
    private function minify($context) {
        //Remove spaces
        $contextMin = preg_replace('#\s+#', ' ', $context);
        //Remove comments
	$contextMin = preg_replace('#/\*.*?\*/#s', '', $contextMin);
        //Remove unuseful spaces
	$contextMin = str_replace('; ', ';', $contextMin);
	$contextMin = str_replace(': ', ':', $contextMin);
	$contextMin = str_replace(' {', '{', $contextMin);
	$contextMin = str_replace('{ ', '{', $contextMin);
	$contextMin = str_replace(', ', ',', $contextMin);
	$contextMin = str_replace('} ', '}', $contextMin);
	$contextMin = str_replace(';}', '}', $contextMin);

        return $contextMin;
    }

    /**
     * Delete directory and his content
     * @param $dir string Directory path
     * @return bool
     */
    private function deleteDirectory($dir) {        
        if (is_dir($dir)) {
            $this->logging("Removing directory '{$dir}' and his content...");
            if (!($objects = scandir($dir))) {
                throw new UnifierException("impossible to open directory '{$dir}'. Check the permissions of this directory.");
            }
            foreach ($objects as $object) {
                if ($object != "." && $object != "..") {
                    if (filetype($dir."/".$object) == "dir") {
                        if (!rrmdir($dir."/".$object)) {
                            throw new UnifierException("impossible to delete directory '{$dir}/{$object}'. Check the permissions of this directory.");
                        }
                    }
                    else {
                        if (!unlink($dir."/".$object)) {
                            throw new UnifierException("impossible to delete file '{$dir}/{$object}'. Check the permissions of this file.");
                        }
                    }
                }
            }
            reset($objects);
            if (!rmdir($dir)) {
                throw new UnifierException("impossible to delete directory '{$dir}'. Check the permissions of this directory.");
            }
        }
        $this->logging("OK!");

        return true;
    }

    /**
     * Append string into logger file
     * @param $line string Text to write
     * @return boolean
     */
    private function logging($line) {
        if ($this->log != null) {
            if (!$fp = fopen($this->log, 'a+')) {
                exit(0);
            }
            chmod($this->log, 0777);

            date_default_timezone_set(@date_default_timezone_get());
            $log =  "*** " . date(DATE_RFC822) . ": " . $line . "\n";
            fwrite($fp, $log, count($log) * 1024);
            fclose($fp);
        }

        return true;
    }

    /**
     * Delete log file
     * @return boolean
     */
    private function deleteLog() {
        if ($this->log != null) {
            if (file_exists($this->log)) {
                unlink($this->log);
            }
        }

        return true;
    }
}

class UnifierException extends Exception {
    public function  __construct($message = null, $code = 0) {
        if (!$message) {
            throw new $this('Unknown '. get_class($this));
        }
        parent::__construct($message, $code);
    }

    public function __toString() {
        return "Error: {$this->message}";
    }
}
?>
